/* Copyright 2000, 2013 Neil Edelman, distributed under the terms of the
 GNU General Public License, see copying.txt */

#include <stdlib.h> /* malloc free */
#include <stdio.h>  /* fprintf */
#include "Open.h"
#include "Entity.h"

/* troll physics */
const static float bounceRatio     = 0.7f;
const static float rotationDamping = 0.8f;
const static float rotationLimit   = 0.0035f;
const static float speedLimit      = 0.03f;
const static float speedZeroFrames = 16.0f;
const static float speedMin        = 0.00003f;

/* from Open.c */
extern struct Open *open;
extern unsigned char (*isTexture)(unsigned);

/* FIXME: "storing 32-bit float result in memory" */

struct Entity {
	struct Vec2f p, v; /* position, velocity */
	float t, w;        /* \theta, \omega */
	struct Vec2f half;
	int texture;
	int stopFrames; /* fixme: msStop instead */
	int isStopped;
};

const static float rand_norm = 1.0f / RAND_MAX;

/* public */

/** constructor */
struct Entity *Entity(const int tex, const int width, const int height) {
	struct Entity *entity;

	if(!open /*|| !isTexture(tex)*/) {
		fprintf(stderr, "Entity: invalid (texture?)\n");
		return 0;
	}
	if(!(entity = malloc(sizeof(struct Entity)))) {
		perror("Entity constructor");
		Entity_(&entity);
		return 0;
	}
	entity->p.x = entity->p.y = 0;
	entity->v.x = entity->v.y = 0;
	entity->t   = entity->w   = 0;
	entity->half.x            = 0.5f * width;
	entity->half.y            = 0.5f * height;
	entity->texture           = tex;
	entity->stopFrames        = 0;
	entity->isStopped         = -1;
	fprintf(stderr, "Entity: new, #%p texture %u (%dx%d.)\n", (void *)entity,
			entity->texture, width, height);

	return entity;
}

/** destructor */
void Entity_(struct Entity **entity_ptr) {
	struct Entity *entity;

	if(!entity_ptr || !(entity = *entity_ptr)) return;
	fprintf(stderr, "~Entity: erase, #%p.\n", (void *)entity);
	free(entity);
	*entity_ptr = entity = 0;
}

int EntityGetTexture(const struct Entity *entity) {
	if(!entity) return 0;
	return entity->texture;
}

#include <math.h>

/** accessor */
/*const struct Vec2f *EntityGetVelocity();*/
struct Vec2f EntityFuturePosition(const struct Entity *entity, const float ms) {
	struct Vec2f p = { 0, 0 };
	if(!entity) return p;
	p.x = entity->p.x + ms * entity->v.x;
	p.y = entity->p.y + ms * entity->v.y;
	return p;
}
float EntityFutureAngle(const struct Entity *entity, const float ms) {
	if(!entity) return 0;
	return entity->t + ms * entity->w;
}
struct Vec2f EntityGetPosition(const struct Entity *entity) {
	struct Vec2f p = { 0, 0 };
	if(!entity) return p;
	p.x = entity->p.x;
	p.y = entity->p.y;
	return p;
}
float EntityGetAngle(const struct Entity *entity) {
	if(!entity) return 0;
	return entity->t;
}
/* fixme: hahaha make my own matrices */
float EntityGetDegrees(const struct Entity *entity) {
	if(!entity) return 0;
	return entity->t * (float)M_180_PI;
}
struct Vec2f EntityGetVelocity(const struct Entity *entity) {
	struct Vec2f v = { 0, 0 };
	if(!entity) return v;
	v.x = entity->v.x;
	v.y = entity->v.y;
	return v;
}
struct Vec2f EntityGetHalf(const struct Entity *entity) {
	struct Vec2f s = { 0, 0 };
	if(!entity) return s;
	s.x = entity->half.x;
	s.y = entity->half.y;
	return s;
}

/** set */
void EntityImpact(struct Entity *entity, const struct Vec2f *impact) {
	if(!entity || !impact) return;
	entity->v.x += impact->x;
	entity->v.y += impact->y;
}

void EntitySetAcceleration(struct Entity *entity, const float a) {
	float t, speed2, correction;
	if(!entity) return;
	t = entity->t /** M_PI / 180.0*/;
	entity->v.x += a * cosf(t); /* fixme?: expesive, lut */
	entity->v.y += a * sinf(t);
	/* enforce the speed limit, allow stopping */
	speed2 = entity->v.x * entity->v.x + entity->v.y * entity->v.y;
	if(entity->isStopped) {
		if(speed2 >= speedMin) entity->isStopped = 0;
	} else {
		if(speed2 > speedLimit) {
			correction = speedLimit / speed2;
			entity->v.x *= correction;
			entity->v.y *= correction;
			/* fprintf(stderr, "Slow down!\n"); <- spam! */
		} else if(speed2 < speedMin) {
			/* should be ms not frames but don't have access, it's not important,
			 I just hate moving always even when you're stopped */
			if(++entity->stopFrames >= speedZeroFrames) {
				entity->v.x = 0;
				entity->v.y = 0;
				entity->isStopped = -1;
				fprintf(stderr, "Stopped.\n"); /* fixme: never happens now */
			}
		} else {
			entity->stopFrames = 0;
		}
	}
}

void EntitySetOmega(struct Entity *entity, const float w) {
	if(!entity) return;
	entity->w += w;
	if(entity->w < -rotationLimit) {
		entity->w = -rotationLimit;
		/* printf("Right turn limit.\n"); <- spam! */
	} else if(entity->w > rotationLimit) {
		entity->w = rotationLimit;
		/* printf("Left turn limit.\n"); <- spam! */
	}
}

void EntityUpdate(struct Entity *entity, const int ms) {
	const struct Vec2f *screen = OpenGetScreen();
	if(!entity) return;
	/* update */
	entity->p.x += entity->v.x * ms;
	entity->p.y += entity->v.y * ms;
	entity->t   += entity->w   * ms;
	/* keep angles b/t [-M_PI - M_PI) for greater precision */
	if(entity->t      < -(float)M_PI) entity->t += (float)M_2PI;
	else if(entity->t >= (float)M_PI) entity->t -= (float)M_2PI;
	/* this should be framerate independant; it's too costly, and the user
	 won't notice 25/26 ms: entity->w *= pow(rotationDamping, ms); */
	/* fixme: framerate dep! */
	entity->w *= rotationDamping;
	/* bounce off the sides */
	if(entity->p.x < -screen->x) {
		/*fprintf(stderr, "Bounce! (%f)\n", -screen->x);*/
		entity->p.x = -screen->x;
		entity->v.x = -entity->v.x * bounceRatio;
	} else if(entity->p.x > screen->x) {
		/*fprintf(stderr, "Bounce! (%f)\n", screen->x);*/
		entity->p.x = screen->x;
		entity->v.x = -entity->v.x * bounceRatio;
	}
	if(entity->p.y < -screen->y) {
		/*fprintf(stderr, "Bounce! (%f)\n", -screen->y);*/
		entity->p.y = -screen->y;
		entity->v.y = -entity->v.y * bounceRatio;
	} else if(entity->p.y > screen->y) {
		/*fprintf(stderr, "Bounce! (%f)\n", screen->y);*/
		entity->p.y = screen->y;
		entity->v.y = -entity->v.y * bounceRatio;
	}
}

void EntityRandomise(struct Entity *entity) {
	const struct Vec2f *screen = OpenGetScreen();
	if(!entity) return;
	entity->p.x = (rand() * 2.0f * rand_norm - 1.0f) * screen->x;
	entity->p.y = (rand() * 2.0f * rand_norm - 1.0f) * screen->y;
	entity->t   = rand() * (float)M_2PI * rand_norm;
}

void EntityPush(struct Entity *entity, float push) {
	if(!entity) return;
	entity->v.x = (rand() * 2.0f * rand_norm - 1.0f) * push;
	entity->v.y = (rand() * 2.0f * rand_norm - 1.0f) * push;
}

/* private */
