/* Copyright 2000, 2013 Neil Edelman, distributed under the terms of the
 GNU General Public License, see copying.txt */

/* OpenGL module; includes graphics and input -Neil */

#include <stdlib.h> /* malloc free */
#include <stdio.h>  /* fprintf */
#include <time.h>   /* srand(clock()) */
#include <math.h>   /* math in shader() */
#ifndef fminf /* c89 */
#define fminf(x, y) (((x) > (y)) ? (x) : (y))
#endif
#ifndef sqrtf /* c89 */
#define sqrtf(x) ((float)sqrt((x)))
#endif
#ifndef atan2f /* c89 */
#define atan2f(x, y) ((float)atan2((x), (y)))
#endif
#include "Glew.h"
#include "Game.h"
#include "Entity.h"
#include "Open.h"
#include "Particle.h"

/* hard coded resouce */
#include "shader.h"
#include "tex_Ngc4038-4039.h"

/* constants */
static const char *programme  = "Void";
static const char *year       = "2000, 2013";
static const int versionMajor = 3;
static const int versionMinor = 1;

static const struct Vec2i screen_init = { 640, 480 };
static const int          time_step   = 20; /* ms */
/* [0 (low = flat), infity (high = conical)] when shading;
 it's sort of like -- . . . /-\ . . . /\ */
static const float conical_shadows    = 2.5f;

static const int key_delay            = 300; /* ms */

/* Key subclass */

struct Key {
	int state;
	int down;
	int integral;
	int time;
};

/* Timer subclass */

#define TIMER_LOG_FRAMES 2

struct Timer {
	int time; /* ms */
	int frame;
	int frameTime[1 << TIMER_LOG_FRAMES];
};

const static int max_frame_time = sizeof(((struct Timer *)0)->frameTime) / sizeof(int);

/* Screen subclass */

struct Screen {
	struct Vec2i dim;
	struct Vec2i saved;
	struct Vec2i savedPos;
	struct Vec2f max;
	int          full;
};

/* class */

struct Open {
	struct Timer  graphics;
	GLenum        lastGlError;
	struct Game   *game;
	struct Key    key[k_max];
	struct Screen screen;
	GLuint        shader;
	GLuint        lighting, background;
	struct Vec2f  inverseBackgroundSize;
	GLuint        vboParticles;
	GLuint        vboParticleColour;
};

/* global! */
struct Open *open = 0;

/* private */
static int isGLError(const char *title);
static void usage(const char *argvz);
static void draw(void);
static void update(int);
static void display(void);
static void background(void);
static void shader(void);
static void shader_(void);

static void Key(struct Key *key);
static enum Keys glut_to_keys(int k);
static void key_down(unsigned char k, int x, int y);
static void key_up(unsigned char k, int x, int y);
static void key_down_special(int k, int x, int y);
static void key_up_special(int k, int x, int y);

static void Timer(struct Timer *timer);
static void advance_time(struct Timer *timer);

static void Screen(struct Screen *screen, const int width, const int height);
static void resize(int width, int height);

/** private (entry point) */
int main(int argc, char **argv) {
	usage(argv[0]);
	srand((unsigned)clock());
	/* negotiate with dynamic library, calculate some address */
	glutInit(&argc, argv);
	/* constructor */
	if(!Open(screen_init.x, screen_init.y, programme)) return EXIT_FAILURE;
	/* loop never returns */
	/* #ifndef GLUT_DISABLE_ATEXIT_HACK */
	if(atexit(&Open_)) perror("~Open");
	glutMainLoop();
	return EXIT_SUCCESS;
}

/** constructor */
struct Open *Open(const int width, const int height, const char *title) {
	int i;

	if(open) return open;
	if(width <= 0 || height <= 0 || !title) {
		fprintf(stderr, "Open: error initialising.\n");
		return 0;
	}
	if(!(open = malloc(sizeof(struct Open)))) {
		perror("Open constructor");
		Open_();
		return 0;
	}
	Timer(&open->graphics);
	open->lastGlError = GL_NO_ERROR;
	open->game        = 0;
	Screen(&open->screen, width, height);
	for(i = 0; i < k_max; i++) Key(&open->key[i]);
	open->shader     = 0;
	open->lighting   = 0;
	open->background = 0;
	open->inverseBackgroundSize.x = open->inverseBackgroundSize.y = 0;
	open->vboParticles = open->vboParticleColour = 0;
	fprintf(stderr, "Open: new, #%p.\n", (void *)open);

	/* create the window */
	glutInitDisplayMode(GLUT_DOUBLE);
	glutInitWindowSize(width, height);
	glutCreateWindow(title);
	fprintf(stderr, "Open: opened GLUT window; vendor %s; version %s; renderer %s; shading language version %s; combined texture image units %d.\n",
		   glGetString(GL_VENDOR), glGetString(GL_VERSION),
		   glGetString(GL_RENDERER), glGetString(GL_SHADING_LANGUAGE_VERSION),
		   GL_MAX_COMBINED_TEXTURE_IMAGE_UNITS /* > 80 */);

	/* set callbacks */
	glutDisplayFunc(&display);
	glutReshapeFunc(&resize);
	glutKeyboardFunc(&key_down);
	glutKeyboardUpFunc(&key_up);
	glutSpecialFunc(&key_down_special);
	glutSpecialUpFunc(&key_up_special);
	/* glutMouseFunc(&mouse); */
	/* glutIdleFunc(0); disable */
	glutTimerFunc(time_step, &update, 0);

	/* enable textures */
	glEnable(GL_TEXTURE_2D);

	/* enable transparency;
	 http://www.opengl.org/sdk/docs/man2/xhtml/glBlendFunc.xml */
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	/* big points */
	glPointSize(5);
	glEnable(GL_POINT_SMOOTH);

	/* call other functions that make this too long */
	background();
	if(!Glew()) {
		fprintf(stderr, "Open: can't use shaders! will start up in OpenGL 1.0 mode.\n");
	} else {
		shader();
	}

	/* set up VBO for particles with the Particle.c */
	fprintf(stderr, "Open: enabling particle vertex buffer object.\n");
	glEnableClientState(GL_VERTEX_ARRAY); /* OGL 1.1 */
#ifdef KEEP_ON_SW /* maybe, I don't really know */
	glVertexPointer(2, GL_FLOAT, 0, ParticleGetPositions());
#else
	glGenBuffers(1, &open->vboParticles);
	glBindBuffer(GL_ARRAY_BUFFER, open->vboParticles);
	glVertexPointer(2, GL_FLOAT, 0, 0); /* OGL 1.1 */
	glBufferData(GL_ARRAY_BUFFER, ParticleGetMax() * sizeof(struct Vec2f), ParticleGetPositions(), GL_STREAM_DRAW);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
#endif
	/* colours */
	glEnableClientState(GL_COLOR_ARRAY); /* OGL 1.1 */
#ifndef KEEP_ON_SW /* no, colour values are . . . different? man, I hate hw */
	glColorPointer(3, GL_FLOAT, 0, ParticleGetColours());
#else
	glGenBuffers(1, &open->vboParticleColour);
	glBindBuffer(GL_ARRAY_BUFFER, open->vboParticleColour);
	glVertexPointer(3, GL_FLOAT, 0, 0); /* OGL 1.1 */
	glBufferData(GL_ARRAY_BUFFER, ParticleGetMax() * sizeof(struct Colour3f), ParticleGetColours(), GL_STREAM_DRAW);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
#endif

	/* get the game set up; must be called after creating an OpenGL context,
	 ie glutCreateWindow */
	if(!(open->game = Game())) {
		Open_();
		return 0;
	}

	/* get the initialisation error code */
	/*fprintf(stderr, "Open: OpenGL initilisation status, \"%s.\"\n", gluErrorString(glGetError()));*/
	isGLError("Open");

	return open;
}

/** destructor */
void Open_(void) {
	if(!open) return;
	Game_(&open->game);
	glDisableClientState(GL_VERTEX_ARRAY/* | GL_COLOR_ARRAY*/);
	if(glIsBuffer(open->vboParticles)) {
		glDeleteBuffers(1, &open->vboParticles);
		fprintf(stderr, "~Open: Deleted particle buffer.\n");
	}
	if(glIsTexture(open->background)) {
		glDeleteTextures(1, &open->background);
		fprintf(stderr, "~Open: Deleted background texture.\n");
	}
	shader_();

	/*fprintf(stderr, "~Open: OpenGL final status \"%s;\" deleting #%p.\n", gluErrorString(glGetError()), (void *)open);*/
	fprintf(stderr, "~Open: deleting OpenGL #%p.\n", (void *)open);
	free(open);
	/* fixme: are you supposed to free textures? (YES!) */
	open = 0;

	isGLError("Open~");
}

/** "static" (include stdarg) */
void OpenPrint(const char *fmt, ...) {
	/*GLUT_BITMAP_8_BY_13
	 GLUT_BITMAP_9_BY_15
	 GLUT_BITMAP_TIMES_ROMAN_10
	 GLUT_BITMAP_TIMES_ROMAN_24
	 GLUT_BITMAP_HELVETICA_10
	 GLUT_BITMAP_HELVETICA_12
	 GLUT_BITMAP_HELVETICA_18 */
	/*va_list ap;*/
	
	if(!fmt || !open) return;
	/* print the chars into the buffer */
	/*va_start(ap, fmt);
	 vsnprintf(open->con, LINE, fmt, ap);
	 va_end(ap);
	 printf("%s\n", open->con);*/
	/* draw overlay */
	glPushMatrix();
	glColor4f(0, 1, 0, 0.5);
	glRasterPos2i(0, 0);
	glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'A');
	glColor4f(1, 1, 1, 1);
	glPopMatrix();

	isGLError("Open::Print");
}

/** "static" toggle fullscreen */
void OpenToggleFullScreen(void) {
	/* TODO: glutGameModeString, glutEnterGameMode looks interesting */
	if(!open) return;
	if(!open->screen.full) {
		fprintf(stderr, "Entering fullscreen.\n");
		open->screen.full = -1;
		open->screen.saved.x = open->screen.dim.x;
		open->screen.saved.y = open->screen.dim.y;
		open->screen.savedPos.x = glutGet(GLUT_WINDOW_X);
		open->screen.savedPos.y = glutGet(GLUT_WINDOW_Y);
		glutFullScreen();
	} else {
		fprintf(stderr, "Exiting fullscreen.\n");
		open->screen.full = 0;
		glutReshapeWindow(open->screen.saved.x, open->screen.saved.y);
		glutPositionWindow(open->screen.savedPos.x, open->screen.savedPos.y);
	}

	isGLError("Open::ToggleFullScreen");
}

/** bmp must be width * height * depth;
 returns a handle to pass to OpenTexture_() */
int OpenTexture(const int width, const int height, const int depth, const unsigned char *bmp) {
	int format = 0, internal = 0;
	int name;

	if(!open || width <= 0 || height <= 0 || !bmp) return 0;
	if(depth == 1) {
		/* we use exclusively for shaders, so I don't think this matters */
		internal = GL_RGB;
		/* format   = GL_LUMINANCE;
		 core context depriciated -- doesn't work on certain versions of OGL
		 maybe this fixes it? I don't know; older Windows versions don't like
		 this or something */
		/*format   = GL_R16;
		format   = GL_R8;*/
		format   = GL_RED;
	} else if(depth == 3) {
		internal = GL_RGB8;
		format   = GL_RGB;
	} else if(depth == 4) {
		internal = GL_RGBA8;
		format   = GL_RGBA;
	} else {
		fprintf(stderr, "Open::Texture: not a reconised depth, %d.\n", depth);
		return 0;
	}
	glGenTextures(1, (unsigned *)&name);
	glBindTexture(GL_TEXTURE_2D, name);
	/* FIXME: OpenGL 2.0? \/ */
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	/* void glTexImage2D(target, level, internalFormat, width, height,
	                     border, format, type, *data); */
	glTexImage2D(GL_TEXTURE_2D, 0, internal, width, height,
				 0, format, GL_UNSIGNED_BYTE, bmp);
	glBindTexture(GL_TEXTURE_2D, 0);
	fprintf(stderr, "Open::Texture: texture %dx%dx%d assigned %u.\n", width, height, depth, name);

	isGLError("Open::Texture");

	return name;
}

void OpenTexture_(int name) {
	if(!name) return;
	fprintf(stderr, "Open::Texture: deleting texture %u.\n", name);
	glDeleteTextures(1, (unsigned int *)&name);

	isGLError("Open::Texture~");
}

/* private */

static void usage(const char *argvz) {
	fprintf(stderr, "Usage: %s\n", argvz);
	fprintf(stderr, "To win, blow up everything that's not you.\n");
	fprintf(stderr, "Player one controls: left, right, up, down, space\n");
	/* fprintf(stderr, "Player two controls: a, d, w, s, space.\n"); */
	fprintf(stderr, "Fullscreen: F1.\n");
	fprintf(stderr, "Exit: Escape.\n\n");
	fprintf(stderr, "Version %d.%d.\n\n", versionMajor, versionMinor);
	fprintf(stderr, "%s Copyright %s Neil Edelman\n", programme, year);
	fprintf(stderr, "This program comes with ABSOLUTELY NO WARRANTY.\n");
	fprintf(stderr, "This is free software, and you are welcome to redistribute it\n");
	fprintf(stderr, "under certain conditions; see copying.txt.\n\n");
}

static int isGLError(const char *title) {
	GLenum err;
	if((err = glGetError()) != GL_NO_ERROR) {
		fprintf(stderr, "%s: OpenGL error: %s.\n", title, gluErrorString(err));
	}
	return (err != GL_NO_ERROR);
}

/* FIXME: vertex array for a square */
/*const static char square_indices[] = { 1, 2, 3, 4 };
const static float square_verts[] = { -8,-8, -8,8, 8,8, 8,-8 };*/

static void draw(void) {
	const int l_no = ParticleGetNo();
	int i;
	float a;
	struct Vec2f v;
	struct Entity *e;

	/* fixme: lighting! */

	/* draw background */
	glMatrixMode(GL_TEXTURE);
	glPushMatrix();
	glTranslatef(0.5, 0.5, 0);
	glScalef(open->screen.max.x * open->inverseBackgroundSize.x * 2.0f, open->screen.max.y * open->inverseBackgroundSize.y * 2.0f, 0);
	glTranslatef(-0.5, -0.5, 0);
	glMatrixMode(GL_MODELVIEW);
	glBindTexture(GL_TEXTURE_2D, open->background);
	glBegin(GL_QUADS);
	glTexCoord2f(0.0f, 0.0f); glVertex2f(-(float)open->screen.max.x * 2.0f, -(float)open->screen.max.y * 2.0f);
	glTexCoord2f(1.0f, 0.0f); glVertex2f((float)open->screen.max.x * 2.0f, -(float)open->screen.max.y * 2.0f);
	glTexCoord2f(1.0f, 1.0f); glVertex2f((float)open->screen.max.x * 2.0f, (float)open->screen.max.y * 2.0f);
	glTexCoord2f(0.0f, 1.0f); glVertex2f(-(float)open->screen.max.x * 2.0f, (float)open->screen.max.y * 2.0f);
	glEnd();
	glMatrixMode(GL_TEXTURE);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);

	/* draw entities */
	if(open->shader) {
		glUseProgram(open->shader);
		glUniform1i(glGetUniformLocation(open->shader, "l_no"), l_no);
		glUniform3fv(glGetUniformLocation(open->shader, "l_lux"), l_no, (float *)ParticleGetColours());
		glUniform2fv(glGetUniformLocation(open->shader, "l_pos"), l_no, (float *)ParticleGetPositions());
	}
	for(i = 0; (e = GameEntity(open->game, i)); i++) {
		v = EntityGetPosition(e);
		a = EntityGetDegrees(e);

		/* shader differentiates b/w unrotated/rotated by the texture matrix */
		if(open->shader) {
			glMatrixMode(GL_TEXTURE);
			glPushMatrix();
			/* fixme: scale(2^0.5) for mismatached rotated/un-rotated! */
			glTranslatef(0.5, 0.5, 0);
			glRotatef(a, 0, 0, 1);
			glTranslatef(-0.5, -0.5, 0);
		}

		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glTranslatef(v.x, v.y, 0);
		glRotatef(a, 0, 0, 1); /* fixme: do my own! */

		glBindTexture(GL_TEXTURE_2D, EntityGetTexture(e));
		/* fixme: vertex array | vbo */
		glBegin(GL_QUADS);
		glTexCoord2f(0.0, 0.0); glVertex2f(-16, -16);
		glTexCoord2f(1.0, 0.0); glVertex2f(16, -16);
		glTexCoord2f(1.0, 1.0); glVertex2f(16, 16);
		glTexCoord2f(0.0, 1.0); glVertex2f(-16, 16);
		glEnd();
		glBindTexture(GL_TEXTURE_2D, 0);
		glPopMatrix();

		if(open->shader) {
			glMatrixMode(GL_TEXTURE);
			glPopMatrix();
		}
	}
	if(open->shader) glUseProgram(0);

	/* draw particles */
	glBindBuffer(GL_ARRAY_BUFFER, open->vboParticles);
	/* faster, but no such fn: glMapBufferRange(GL_ARRAY_BUFFER, 0, ParticleGetNo(), GL_MAP_WRITE_BIT | GL_MAP_UNSYNCHRONIZED_BIT);*/
#ifndef KEEP_ON_SW
	glBufferSubData(GL_ARRAY_BUFFER, 0, ParticleGetNo() * sizeof(struct Vec2f), ParticleGetPositions());
	/*glBufferSubData(GL_ARRAY_BUFFER, 0, ParticleGetNo() * sizeof(struct Colour3f), ParticleGetColours());*/
#endif
	glDrawArrays(GL_POINTS, 0, ParticleGetNo());
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	/********* fixed. FIXME! vertex() crashes Windows ***********/
	/* glPointSize(3); glBegin(GL_POINTS); ParticleDraw(); glEnd(); glColor4f(1, 1, 1, 1); */
}

/** callback for glutTimerFunc */
static void update(int value) {
	glutTimerFunc(time_step, update, 0);
	/* todo: have graphics and logic timers instead of trusting time_step */
	GameUpdate(open->game, time_step);
	glutPostRedisplay();
}

/** callback for glutDisplayFunc */
static void display(void) {
	/* GLenum err; */

	/* has a background -> glClear(GL_COLOR_BUFFER_BIT);*/
	advance_time(&open->graphics);
	/* draw */
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	draw();
	/* swap */
	glutSwapBuffers();
	/* check for errors */
	isGLError("Open::display");
	/*if((err = glGetError()) != GL_NO_ERROR && err != open->lastGlError) {
		fprintf(stderr, "Open::display: OpenGL error \"%s.\"\n", gluErrorString(err));
		open->lastGlError = err;
	}*/
}

/** this is the background, called only in Open constructor
 fixme: OpenGL < 2.0 requires square textures; don't know what happens! */
static void background(void) {
	int name;

	/* max texture >= 1024 */
	glGenTextures(1, (unsigned *)&name);
	glBindTexture(GL_TEXTURE_2D, name);
	/* fixme: what happens if we have less than opengl 2.0 with the params?
	 I believe what should happen is it ignores it and it's still GL_REPEAT */
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_MIRRORED_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_MIRRORED_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	/* void glTexImage2D(target, level, internalFormat, width, height,
	 border, format, type, *data); */
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB8, Ngc__________bmp_width, Ngc__________bmp_height,
				 0, GL_RGB, GL_UNSIGNED_BYTE, Ngc__________bmp_bmp);
	open->background = name;
	open->inverseBackgroundSize.x = 2.0f / Ngc__________bmp_width;
	open->inverseBackgroundSize.y = 2.0f / Ngc__________bmp_height;
	fprintf(stderr, "Activated background texture %d x %d.\n", Ngc__________bmp_width, Ngc__________bmp_height);

	isGLError("Open::background");
}

/** creates the shadows by compiling shader programme,
 called only in Open constructor */
static void shader(void) {
	GLuint vert, frag;
	GLint status;
	GLchar info[128];

	/* texture that have to be there for the shader to work */
	glActiveTexture(GL_TEXTURE1);
	{
		int i, j;
		float x, y;
		/* variable [ysize][xsize] of the lighting sprite, must be 2 'colors' */
		float buffer[64][64][2];
		const int buffer_ysize = sizeof(buffer)  / sizeof(*buffer);
		const int buffer_xsize = sizeof(*buffer) / sizeof(**buffer);
		const float buffer_norm = conical_shadows * 2.0f / sqrtf(buffer_xsize * buffer_xsize + buffer_ysize * buffer_ysize);
		int name;

		for(j = 0; j < buffer_ysize; j++) {
			for(i = 0; i < buffer_xsize; i++) {
				x = (float)i - 0.5f*buffer_xsize + 0.5f;
				y = (float)j - 0.5f*buffer_ysize + 0.5f;
				buffer[j][i][0] = fminf(sqrtf(x*x + y*y) * buffer_norm, 1.0f);
				/* NOTE: opengl clips [0, 1), even if it says different;
				 maybe it's GL_LINEAR? */
				buffer[j][i][1] = fmodf(atan2f(y, x) + M_2PI, M_2PI) * (float)M_1_2PI;
			}
		}
		glGenTextures(1, (unsigned *)&name);
		glBindTexture(GL_TEXTURE_2D, name);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		/* NOTE: GL_LINEAR is a lot prettier, but causes a crease when the
		 principal value goes from [0, 1) ([0, 2Pi)) */
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		/* void glTexImage2D(target, level, internalFormat, width, height,
		 border, format, type, *data); */
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RG32F, buffer_xsize, buffer_ysize,
					 0, GL_RG, GL_FLOAT, buffer);
		open->lighting = name;
		fprintf(stderr, "Activated lighting texture.\n");
	}
	glActiveTexture(GL_TEXTURE0);
	/* create shaders; the 0 at the end is null-terminated */
	vert = glCreateShader(GL_VERTEX_SHADER);
	glShaderSource(vert, 1, &shader_vs, 0);
	frag = glCreateShader(GL_FRAGMENT_SHADER);
	glShaderSource(frag, 1, &shader_fs, 0);
	/* compile vertex shader */
	glCompileShader(vert);
	glGetShaderiv(vert, GL_COMPILE_STATUS, &status);
	if(!status) {
		glGetShaderInfoLog(vert, sizeof(info), 0, info);
		fprintf(stderr, "Failed compiling vertex shader:\n");
		if(*info) fprintf(stderr, "%s\n", info);
		glDeleteShader(vert);
		glDeleteShader(frag);
		shader_();
		return;
	} else {
		fprintf(stderr, "Successfully compiled vertex shader.\n");
	}
	/* compile fragment shader */
	glCompileShader(frag);
	glGetShaderiv(frag, GL_COMPILE_STATUS, &status);
	if(!status) {
		glGetShaderInfoLog(frag, sizeof(info), 0, info);
		fprintf(stderr, "Failed compiling fragment shader:\n");
		if(*info) fprintf(stderr, "%s\n", info);
		glDeleteShader(vert);
		glDeleteShader(frag);
		shader_();
		return;
	} else {
		fprintf(stderr, "Successfully compiled fragment shader.\n");
	}
	/* link */
	open->shader = glCreateProgram();
	glAttachShader(open->shader, vert);
	glAttachShader(open->shader, frag);
	glLinkProgram(open->shader);
	glDetachShader(open->shader, frag);
	glDetachShader(open->shader, vert);
	glDeleteShader(vert);
	glDeleteShader(frag);
	/* success? */
	glGetProgramiv(open->shader, GL_LINK_STATUS, &status);
	if(!status) {
		glGetProgramInfoLog(open->shader, sizeof(info), 0, info);
		fprintf(stderr, "Failed linking shader programme:\n");
		if(*info) fprintf(stderr, "%s\n", info);
		shader_();
		return;
	} else {
		fprintf(stderr, "Successfully linked shader programme.\n");
	}
	/* check for OpenGL's ability to execute this programme */
	glValidateProgram(open->shader);
	glGetProgramiv(open->shader, GL_VALIDATE_STATUS, &status);
	if(!status) {
		fprintf(stderr, "Failed validating shader programme.\n");
		shader_();
		return;
	} else {
		fprintf(stderr, "Successfully validated shader programme.\n");
	}
	/* constant values (the GL_TEXTUREx) */
	glUseProgram(open->shader);
	glUniform1i(glGetUniformLocation(open->shader, "sprite"), 0);
	glUniform1i(glGetUniformLocation(open->shader, "lighting"), 1);
	glUseProgram(0);

	isGLError("Open::shader");
}

static void shader_(void) {
	if(open->shader) {
		glDeleteProgram(open->shader);
		open->shader = 0;
		fprintf(stderr, "Unlinked fragment shader programme.\n");
	}
	if(glIsTexture(open->lighting)) {
		glDeleteTextures(1, &open->lighting);
		fprintf(stderr, "Deleted lighting texture.\n");
	}

	isGLError("Open::shader!");
}

/************************************************/

/** subclass Key */
static void Key(struct Key *key) {
	if(!key) return;
	/* all keys are turned off;
	 that's not necesarily true, but user expects this */
	key->state    = 0;
	key->down     = 0;
	key->integral = 0;
	key->time     = 0;
}

/** "static" check for how long the keys has been pressed, without repeat rate
 (destructive) */
int OpenKeyTime(const int key) {
	int time;
	struct Key *k;
	if(key < 0 || key >= k_max || !open) return 0;
	k = &open->key[key];
	if(k->state) {
		time    = open->graphics.time - k->down + k->integral;
		k->down = open->graphics.time;
	} else {
		time    = k->integral;
	}
	k->integral = 0;
	return time;
}

/** "static" key press, with repeat rate (destructive) */ 
int OpenKeyPress(const int key) {
	int time;
	struct Key *k;
	if(key < 0 || key >= k_max || !open) return 0;
	k = &open->key[key];
	if(k->state || k->integral) {
		k->integral = 0;
		time = glutGet(GLUT_ELAPSED_TIME);
		if(time > k->time + key_delay) {
			k->time = time;
			return -1;
		}
	}
	return 0;
}

/** GLUT_ to internal keys (are you sure that this is not standard?) */
static enum Keys glut_to_keys(int k) {
	switch(k) {
		case GLUT_KEY_LEFT:   return k_left;
		case GLUT_KEY_UP:     return k_up;
		case GLUT_KEY_RIGHT:  return k_right;
		case GLUT_KEY_DOWN:   return k_down;
		case GLUT_KEY_PAGE_UP:return k_pgup;
		case GLUT_KEY_PAGE_DOWN: return k_pgdn;
		case GLUT_KEY_HOME:   return k_home;
		case GLUT_KEY_END:    return k_end;
		case GLUT_KEY_INSERT: return k_ins;
		case GLUT_KEY_F1:     return k_f1;
		case GLUT_KEY_F2:     return k_f2;
		case GLUT_KEY_F3:     return k_f3;
		case GLUT_KEY_F4:     return k_f4;
		case GLUT_KEY_F5:     return k_f5;
		case GLUT_KEY_F6:     return k_f6;
		case GLUT_KEY_F7:     return k_f7;
		case GLUT_KEY_F8:     return k_f8;
		case GLUT_KEY_F9:     return k_f9;
		case GLUT_KEY_F10:    return k_f10;
		case GLUT_KEY_F11:    return k_f11;
		case GLUT_KEY_F12:    return k_f12;
		default: return k_unknown;
	}
}

/** callback for glutKeyboardFunc */
static void key_down(unsigned char k, int x, int y) {
	struct Key *key = &open->key[k];
	if(key->state) return;
	key->state  = -1;
	key->down = glutGet(GLUT_ELAPSED_TIME);
	/* fprintf(stderr, "Open::key_down: key %d hit at %d ms.\n", k, key->down); */
}

/** callback for glutKeyboardUpFunc */
static void key_up(unsigned char k, int x, int y) {
	struct Key *key = &open->key[k];
	if(!key->state) return;
	key->state = 0;
	key->integral += glutGet(GLUT_ELAPSED_TIME) - key->down;
	/* fprintf(stderr, "Open::key_up: key %d pressed %d ms at end of frame.\n", k, key->integral); */
}

/** callback for glutSpecialFunc */
static void key_down_special(int k, int x, int y) {
	struct Key *key = &open->key[glut_to_keys(k)];
	if(key->state) return;
	key->state  = -1;
	key->down = glutGet(GLUT_ELAPSED_TIME);
	/* fprintf(stderr, "Open::key_down_special: key %d hit at %d ms.\n", k, key->down); */
}

/** callback for glutSpecialUpFunc */
static void key_up_special(int k, int x, int y) {
	struct Key *key = &open->key[glut_to_keys(k)];
	if(!key->state) return;
	key->state = 0;
	key->integral += glutGet(GLUT_ELAPSED_TIME) - key->down;
	/* fprintf(stderr, "Open::key_up_special: key %d pressed %d ms at end of frame.\n", k, key->integral); */
}

/************************************************/

/** subclass Timer */
static void Timer(struct Timer *timer) {
	int i;

	timer->time        = glutGet(GLUT_ELAPSED_TIME);
	timer->frame       = 0;
	for(i = 0; i < max_frame_time; i++) timer->frameTime[i] = timer->time;
}

/** advances the timer every eg frame, logic step */
static void advance_time(struct Timer *timer) {
	timer->time  = glutGet(GLUT_ELAPSED_TIME);
	timer->frame = (timer->frame + 1) & (max_frame_time - 1);
	timer->frameTime[timer->frame] = timer->time;	
}

/************************************************/

/** subclass Screen */
static void Screen(struct Screen *screen, const int width, const int height) {
	screen->dim.x      = width;
	screen->dim.y      = height;
	screen->max.x      = 0.5f * screen->dim.x;
	screen->max.y      = 0.5f * screen->dim.y;	
	screen->saved.x    = screen->dim.x;
	screen->saved.y    = screen->dim.y;
	screen->savedPos.x = 0;
	screen->savedPos.y = 0;
	screen->full       = 0;
}

/** query screen hight/width or get default if no open; must never return null */
const struct Vec2f *OpenGetScreen(void) {
	if(!open) return 0;
	return &open->screen.max;
}

/** callback for glutReshapeFunc */
static void resize(int width, int height) {
	float half_width = 0.5f * width, half_height = 0.5f * height;
	if(width <= 0 || height <= 0) return;
	glViewport(0, 0, width, height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(-half_width, half_width, -half_height, half_height);
	if(!open) return;
	open->screen.dim.x = width;
	open->screen.dim.y = height;
	open->screen.max.x = half_width;
	open->screen.max.y = half_height;

	isGLError("Open::resize");
}
