/* Copyright 2000, 2013 Neil Edelman, distributed under the terms of the
 GNU General Public License, see copying.txt */

#include <stdlib.h> /* malloc free */
#include <stdio.h>  /* fprintf */
#include "Open.h"
#include "Entity.h" /* for ParticleCollide() */
#include "Particle.h"

/* physics */

/* from Open.c */
extern struct Open *open;
extern void (*vertex)(float, float);
extern void (*colour)(float, float, float, float);

/* private */

#define MAX_LIGHTS 64 /* has to be the same as in the shader */

static const float particle_alpha = 0.8f;

/* FIXME: "storing 32-bit float result in memory" */

struct Particle {
	struct Vec2f v;
	int damage;  /* 0, intangable */
	int expires; /* 0, no expiry */
	char name;
	/* fixme: callback to have homing */
	/* fixme: place in order to have lighting! */
	/*struct Particle *xNext;*/
};

/* lights, separate these values in a contiguous array for shaders, really a
 part of the structure above! */
int                    l_no;
struct Colour3f        l_lux[MAX_LIGHTS];
struct Vec2f           l_pos[MAX_LIGHTS];
static struct Particle particle[MAX_LIGHTS];
const static int maxParticles = sizeof(particle) / sizeof(struct Particle);

/* public */

/** constructor */
int Particle(const struct Vec2f x, const struct Vec2f v, const struct Colour3f lux, const int damage, const int expires) {
	const struct Vec2f *screen = OpenGetScreen();
	int no;
	struct Particle *p;
	static char name = 'A';

	if(!open) {
		fprintf(stderr, "Particle: graphics not up.\n");
		return -1;
	}
	if(x.x < -screen->x || x.x > screen->x || x.y < -screen->y || x.y > screen->y) {
		fprintf(stderr, "Particle: not created offscreen.\n");
		return -1;
	}
	if(l_no >= MAX_LIGHTS/*noParticles >= maxParticles*/) {
		fprintf(stderr, "Particle: reached limit of %d/%d lights.\n", l_no, MAX_LIGHTS);
		return -1;
	}
	no = l_no++;
	p  = &particle[no];
	l_pos[no].x = x.x;
	l_pos[no].y = x.y;
	p->v.x = v.x;
	p->v.y = v.y;
	l_lux[no].r = lux.r;
	l_lux[no].g = lux.g;
	l_lux[no].b = lux.b;
	/*l_lux[no].a = particle_alpha;*/
	p->damage  = damage;
	p->expires = expires;
	p->name    = name++;
	if(name == 'z') name = 'A';
	/*fprintf(stderr, "Particle: new colour (%f,%f,%f) at (%f,%f), #%d.\n",
			lux.r, lux.g, lux.b, x.x, x.y, no); <- spam! */
	/*fprintf(stderr, "Particle: new '%c', #%d.\n", p->name, no);
	{
		int i;
		for(i = 0; i < l_no; i++) fprintf(stderr, "%c, ", particle[i].name);
		fprintf(stderr, "0.\n");
	}*/

	return no;
}

/** destructor of no, will update with new particle at no */
void Particle_(const int no) {
	if(no < 0 || no >= l_no) return;
	/* fprintf(stderr, "~Particle: erase, #%d.\n", no); <- spam! */
	/*fprintf(stderr, "~Particle: erase '%c', #%d.\n", particle[no].name, no);*/
	/* place the noParticles item into this one, decrese #; nobody will know */
	if(no < l_no - 1) {
		struct Particle *p, *z;
		p = &particle[no];
		z = &particle[l_no - 1];
		l_pos[no].x = l_pos[l_no - 1].x;
		l_pos[no].y = l_pos[l_no - 1].y;
		p->v.x = z->v.x;
		p->v.y = z->v.y;
		l_lux[no].r = l_lux[l_no - 1].r;
		l_lux[no].g = l_lux[l_no - 1].g;
		l_lux[no].b = l_lux[l_no - 1].b;
		/*l_lux[no].a = l_lux[l_no - 1].a;*/
		p->damage  = z->damage;
		p->expires = z->expires;
		p->name    = z->name;
	}
	l_no--;
	/*{
		int i;
		for(i = 0; i < noParticles; i++) fprintf(stderr, "%c, ", particle[i].name);
		fprintf(stderr, "0.\n");
	}*/
}

/** get address of things for shaders and VBOs in Open.c */
int ParticleGetNo(void) { return l_no; }
const struct Colour3f *ParticleGetColours(void) { return l_lux; }
const struct Vec2f *ParticleGetPositions(void) { return l_pos; }
int ParticleGetMax(void) { return MAX_LIGHTS; }

/** draws the particles; NOTE: not working in Windows, replaced by VBO */
/*void ParticleDraw(void) {
	int i;
	struct Particle *p;
	for(i = 0, p = particle; i < l_no; i++, p++) {
		colour(l_lux[i].r, l_lux[i].g, l_lux[i].b, particle_alpha);
		vertex(l_pos[i].x, l_pos[i].y);
	}
}*/

void ParticleUpdate(const int ms) {
	const struct Vec2f *screen = OpenGetScreen();
	int i;
	struct Particle *p;
	for(i = 0, p = particle; i < l_no; ) {
		/* expire */
		if(p->expires && (p->expires -= ms) <= 0) {
			Particle_(i);
			continue;
		}
		l_pos[i].x += p->v.x * ms;
		l_pos[i].y += p->v.y * ms;
		if(l_pos[i].x < -screen->x
		   || l_pos[i].x > screen->x
		   || l_pos[i].y < -screen->y
		   || l_pos[i].y > screen->y) {
			Particle_(i);
			continue; /* will update with new particle */
		}
		/* fixme: do collision detection if damage > 0 */
		i++;
		p++;
	}
}

/* do collision detection with shots on screen; return how much damage
 and out *push */
/* fixme: improve this by checking the entire path and not just the end */
/* fixme: check further if the part of the ship hit is opaque */
/* fixme: make a doexplosion func to just push everything around out and
 seed an explosion graphic */
int ParticleCollide(const struct Entity *e, struct Vec2f *push) {
	int i, damage = 0;
	struct Particle *p;
	struct Vec2f pos, size;

	if(!push) return 0;
	push->x = 0;
	push->y = 0;
	if(!e) return 0;
	pos  = EntityGetPosition(e);
	size = EntityGetHalf(e);
	for(i = 0, p = particle; i < l_no; ) {
		if(p->damage
		   && l_pos[i].x >= pos.x - size.x
		   && l_pos[i].x <= pos.x + size.x
		   && l_pos[i].y >= pos.y - size.y
		   && l_pos[i].y <= pos.y + size.y) {
			damage += p->damage;
			push->x += p->v.x;
			push->y += p->v.y;
			Particle_(i);
			continue; /* will update with new particle */
		}
		i++;
		p++;
	}
	return damage;
}
